import { system, world, EquipmentSlot, EntityEquippableComponent, Effect, EffectType } from "@minecraft/server";

// #region Utils
import { getObjective, getEquipment, getPlayers, classTests, cooldownMap, commandMap, standardMap, displayMap, chargeMap, shieldChargeLocked, } from "./combat_utils.js";
// #endregion
const scoreboardWeaponClass = getObjective("htc_weapon_class");
const scoreboardCooldown = getObjective("htc_attack_cooldown");
const scoreboardParryCooldown = getObjective("htc_parry_cooldown");
const scoreboardShieldCharge = getObjective("htc_shield_charge");
const scoreboardShieldCharging = getObjective("htc_shield_charging");
const inanimateQuery = {excludeFamilies: ["inanimate", "npc"], excludeTypes: ["item", "painting", "ender_dragon", "arrow", "thrown_trident", "snowball", "egg", "ender_pearl", "eye_of_ender_signal", "fishing_hook"] };



system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const playerEquipment = getEquipment(player);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        let weaponClassValue;
        try { weaponClassValue = scoreboardWeaponClass.getScore(player); } catch { weaponClassValue = 0; scoreboardWeaponClass.setScore(player, 0); }
        if (mainHandItem) {
            if (!classTests.some(classTest => classTest.some(typeId => mainHandItem.typeId.includes(typeId)))) {
                scoreboardWeaponClass.setScore(player, 0);
                player.runCommand("title @a times 0 3 3")
            }
            classTests.forEach((classTest, index) => {
                if (classTest.find(typeId => mainHandItem.typeId.includes(typeId)) && (weaponClassValue !== index + 1 || weaponClassValue === undefined)) {
                    scoreboardWeaponClass.setScore(player, index + 1);
                }
            });
        } else { scoreboardWeaponClass.setScore(player, 0); }
}},1);

world.afterEvents.entityHitEntity.subscribe((evd) => {
    const damagingEntity = evd.damagingEntity;
    const hitEntity = evd.hitEntity;
    if (damagingEntity.typeId !== "minecraft:player") return;
    if (damagingEntity && hitEntity) {
        const cooldownValue = scoreboardCooldown.getScore(damagingEntity);
        const weaponClassValue = scoreboardWeaponClass.getScore(damagingEntity);
        const isOnGroundAndNotFalling = damagingEntity.isOnGround && !damagingEntity.isFalling;
        const isInWater = damagingEntity.isInWater
        const command = commandMap[weaponClassValue];
        
        const playerViewDirection = damagingEntity.getViewDirection();
        const playerDirectionX = playerViewDirection.x;
        const playerDirectionY = playerViewDirection.y;
        const playerDirectionZ = playerViewDirection.z;

        const playerEquipment = damagingEntity.getComponent(EntityEquippableComponent.componentId);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        if (weaponClassValue in cooldownMap) {
            const newCooldownValue = cooldownMap[weaponClassValue];
            if ((cooldownValue <= newCooldownValue) || (cooldownValue === undefined)) { scoreboardCooldown.setScore(damagingEntity, newCooldownValue); }
            if ((cooldownValue === undefined || cooldownValue <= 0) && isOnGroundAndNotFalling && standardMap.includes(weaponClassValue)) {
                damagingEntity.runCommand(`function ${command}`);
            }
            if (mainHandItem) {
                if ((mainHandItem.typeId.includes("shield") || mainHandItem.typeId.includes("slime_") || mainHandItem.typeId.includes("_nunchaku"))) {
                    if (mainHandItem.typeId.includes("shield") && (cooldownValue === undefined || cooldownValue <= 0)) {
                        damagingEntity.dimension.getEntities({ ...inanimateQuery, minDistance:0 , maxDistance: 1.4, location: hitEntity.location }).forEach(e => {
                            if (e == damagingEntity) return;
                            e.applyKnockback(playerDirectionX, playerDirectionZ, 1.2, 0.1);
                            e.applyDamage(3, { cause: "entityAttack", damagingEntity: damagingEntity });
                        });
                    }
                    if (mainHandItem.typeId.includes("slime_") && !mainHandItem.typeId.includes("_nunchaku")) { hitEntity.applyKnockback(playerDirectionX, playerDirectionZ, 1.4, 0.25); }
                    if (mainHandItem.typeId.includes("_nunchaku") && !mainHandItem.typeId.includes("slime_")) { hitEntity.applyKnockback(playerDirectionX, playerDirectionZ, 0.25, 0.25); }
                    if (mainHandItem.typeId.includes("slime_") && mainHandItem.typeId.includes("_nunchaku")) { hitEntity.applyKnockback(playerDirectionX, playerDirectionZ, 0.4, 0.4); }
                }
            }
            if (weaponClassValue === 11) {
                if (!isOnGroundAndNotFalling && !isInWater && playerDirectionY < 0) { damagingEntity.applyKnockback(playerDirectionX, playerDirectionZ, 0, 0.8);
                } else if (!isInWater) { damagingEntity.applyKnockback(playerDirectionX, playerDirectionZ, -1.3, 0) }
            }
            if (weaponClassValue === 12) {
                if (!isOnGroundAndNotFalling && !isInWater) { damagingEntity.applyKnockback(playerDirectionX, playerDirectionZ, -1.3, -0.6);
                } else if (!isInWater) { damagingEntity.applyKnockback(playerDirectionX, playerDirectionZ, 0.8, 0.0) }
            }
            if ((cooldownValue === undefined || cooldownValue <= 0) && !isOnGroundAndNotFalling && !isInWater && weaponClassValue === 13) {
                damagingEntity.runCommand(`function ${command}`);
                damagingEntity.playSound("mob.wither.break_block", { location:hitEntity.location, pitch:0.8, volume:0.1 })
                damagingEntity.dimension.getEntities({ ...inanimateQuery, minDistance:0.5 , maxDistance: 3, location: damagingEntity.location }).forEach(e => { e.applyKnockback(playerDirectionX, playerDirectionZ, 0, 0.8)});
            }
            /*if (weaponClassValue === 14) {
                const entityHealth = hitEntity.getComponent("minecraft:health");
                const entityMaxHealth = hitEntity.getComponent("minecraft:health").effectiveMax;
                //const beheadingChance = Math.floor(Math.random()*6);
                if (entityHealth <= ((entityMaxHealth/100)*33) && (hitEntity.typeId !== "minecraft:ender_dragon")) {
                    if (hitEntity.typeId === "minecraft:creeper") { hitEntity.runCommand(`loot spawn ~~~ loot creeper_decap`); }
                    if (hitEntity.typeId === "minecraft:skeleton") { hitEntity.runCommand(`loot spawn ~~~ loot skeleton_decap`); }
                    if (hitEntity.typeId === "minecraft:wither_skeleton") { hitEntity.runCommand(`loot spawn ~~~ loot wither_skeleton_decap`); }
                    if (hitEntity.typeId === "minecraft:zombie") { hitEntity.runCommand(`loot spawn ~~~ loot zombie_decap`); }
                    hitEntity.applyDamage((entityHealth), {cause:"override", damagingEntity: damagingEntity});
                } else if (entityHealth <= ((entityMaxHealth/100)*16) && (hitEntity.typeId === "minecraft:ender_dragon")) {
                    hitEntity.runCommand(`loot spawn ~~~ loot dragon_decap`);
                    hitEntity.applyDamage((entityHealth), {cause:"override", damagingEntity: damagingEntity});
                }
            }*/
        }
}});

world.afterEvents.playerSpawn.subscribe((event) => {
    event.player.onScreenDisplay.setTitle('', {
        stayDuration: 36,
        fadeInDuration: 8,
        fadeOutDuration: 20,
        subtitle: '',
    });
});

system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const cooldownValue = scoreboardCooldown.getScore(player);
        let cooldown = displayMap[cooldownValue] ?? "";
        if (cooldownValue > 0) {
            player.onScreenDisplay.setTitle('', { stayDuration: 2, fadeInDuration: 0, fadeOutDuration: 2, subtitle: '', });
            if (cooldownValue > 1) { player.onScreenDisplay.updateSubtitle(cooldown);
                if (cooldownValue > 3) { player.addEffect("weakness", 2, {amplifier:0, showParticles:false});
                    if (cooldownValue >= 9) { player.addEffect("weakness", 2, {amplifier:1, showParticles:false});
                        if (cooldownValue >= 15) { player.addEffect("weakness", 2, {amplifier:2, showParticles:false}); player.addEffect("mining_fatigue", 1, {amplifier:2, showParticles:false});
                            if (cooldownValue >= 21) { player.addEffect("weakness", 2, {amplifier:3, showParticles:false}); }
            }}}} else if (cooldownValue === 1) { player.onScreenDisplay.setTitle('', { stayDuration: 8, fadeInDuration: 0, fadeOutDuration: 4, subtitle: '', }); }
            scoreboardCooldown.addScore(player, -1);
        }
}});

system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const twoHandedMap = ["minecraft:mace", "_battleaxe","_claymore","_greatsword","_hammer","_katana","_lance","_longsword","_morningstar","_scythe"];
        const twoHandedDynamic = "hax:two_handed_weapon";
        const playerEquipment = getEquipment(player);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        if (mainHandItem) {
            if (twoHandedMap.some(typeId => mainHandItem.typeId.includes(typeId))) { player.setDynamicProperty(twoHandedDynamic, 1); } 
            else { player.setDynamicProperty(twoHandedDynamic, 0); }
        } else { player.setDynamicProperty(twoHandedDynamic, 0); }
}},1);

const offhandIgnored = ["map"];
system.runInterval(() => {
  for (const player of world.getPlayers()) {
    const offhandFullDynamic = "hax:offhand_full";
    const playerEquipment = getEquipment(player);
    const offHandItem = playerEquipment?.getEquipment(EquipmentSlot.Offhand);
    // Set to false if no item or if the item is ignored
    if (!offHandItem || offhandIgnored.some((typeId) => offHandItem.typeId.includes(typeId))) {
      player.setDynamicProperty(offhandFullDynamic, false);
    } else {
      player.setDynamicProperty(offhandFullDynamic, true);
    }
  }
}, 4);

system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const twoHandedValue = player?.getDynamicProperty('hax:two_handed_weapon');
        const offhandFullValue = player?.getDynamicProperty('hax:offhand_full');
        if ((twoHandedValue === 1) && (offhandFullValue) && !player.getEffect('weakness')) {
            player.addEffect("weakness", 8, {amplifier:0, showParticles:false})
        }
}});

const parryTest = ["_claymore","_greatsword","_hammer","_katana","_lance","_longsword","_rapier","_sai"];
system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const playerEquipment = getEquipment(player);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        const parryCooldownValue = scoreboardParryCooldown.getScore(player)
        const isParryOverOrUndefined = parryCooldownValue === undefined || parryCooldownValue <= -20;
        if (mainHandItem) {
            if (parryTest.some(typeId => mainHandItem.typeId.includes(typeId)) && player.isSneaking && isParryOverOrUndefined) {
                scoreboardParryCooldown.setScore(player, 7);
                player.addEffect("resistance", 6, {amplifier:2, showParticles:false})
            }
        }
        if (parryCooldownValue > -20) { scoreboardParryCooldown.addScore(player, -1); } else if (parryCooldownValue <= -20) { scoreboardParryCooldown.removeParticipant(player) }
}});

// #region Shield Charge
system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const playerEquipment = getEquipment(player);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        const shieldChargeValue = scoreboardShieldCharge.getScore(player)??0
        const chargeLockedValue = player?.getDynamicProperty(shieldChargeLocked);

        const isOnGroundAndNotFalling = player.isOnGround && !player.isFalling;
        let chargeMeter = chargeMap[shieldChargeValue] ?? "";
        if (mainHandItem) {
            if (mainHandItem.typeId.includes("shield")) {
                if (player.isSneaking && isOnGroundAndNotFalling && !chargeLockedValue && shieldChargeValue <= 12) {
                    player.onScreenDisplay.setActionBar(chargeMeter);
                    if (shieldChargeValue < 12) { scoreboardShieldCharge.addScore(player, 1); }
                }
                if (!player.isSneaking && (shieldChargeValue >= 1)) {
                    if (!player.isSneaking && (shieldChargeValue <= 4)) { scoreboardShieldCharge.removeParticipant(player);
                    } else { player.setDynamicProperty(shieldChargeLocked, true); }
                }
            } else if (shieldChargeValue <= 4) { scoreboardShieldCharge.removeParticipant(player); }
        }
}},5);

system.runInterval(() => {
    for (const player of world.getPlayers()) {
        const playerEquipment = getEquipment(player);
        const mainHandItem = playerEquipment?.getEquipment(EquipmentSlot.Mainhand);
        const shieldChargeValue = scoreboardShieldCharge.getScore(player)
        const chargeLockedValue = player?.getDynamicProperty(shieldChargeLocked);

        const playerViewDirection = player.getViewDirection();
        const playerDirectionX = playerViewDirection.x;
        const playerDirectionZ = playerViewDirection.z;
        const isOnGroundAndNotFalling = player.isOnGround && !player.isFalling;
        if (mainHandItem) {
            const shieldChargingValue = scoreboardShieldCharging.getScore(player)
            if (mainHandItem.typeId.includes("shield")) {
                if (player.isSneaking && (chargeLockedValue)) {
                    if (isOnGroundAndNotFalling) { player.applyKnockback(playerDirectionX, playerDirectionZ, (0.5*shieldChargeValue), 0.3); scoreboardShieldCharging.setScore(player, ((shieldChargeValue-4))*2);
                    } else { player.applyKnockback(playerDirectionX, playerDirectionZ, (0.4*shieldChargeValue), 0.1); }
                    scoreboardShieldCharge.removeParticipant(player);
                    player.setDynamicProperty(shieldChargeLocked, false);
                }
                if ((shieldChargingValue > 0)) {
                    scoreboardShieldCharging.addScore(player, -1);
                    player.dimension.getEntities({ ...inanimateQuery, minDistance:0 , maxDistance: 1.4, location: player.location }).forEach(e => {
                            if (e == player) return;
                            e.applyKnockback(playerDirectionX, playerDirectionZ, (shieldChargingValue/12), (shieldChargingValue/32));
                            e.applyDamage((Math.log(shieldChargingValue)*3.25), { cause: "charging", damagingEntity: player });
                    });
                }
            }
        }
}});
// #endregion

world.afterEvents.entityHurt.subscribe(({damage, damageSource, hurtEntity}) => {
    const damageCauseTypesFilter = ["entityAttack", "entityExplosion", "fall", "fireworks", "flyIntoWall", "projectile", "ramAttack"]
    const parryCooldownValue = scoreboardParryCooldown.getScore(hurtEntity)
    if (12 > parryCooldownValue > 0) {
        if(damageCauseTypesFilter.some((damageType) => damageSource.cause == damageType)) {
            scoreboardParryCooldown.removeParticipant(hurtEntity);
            hurtEntity.addEffect("strength", 14, {amplifier:0, showParticles:false})
            hurtEntity.playSound("block.bell.hit", { location:hurtEntity.location, pitch:3.8, volume:0.2 })
            hurtEntity.runCommand(`execute as @s anchored eyes run particle minecraft:critical_hit_emitter`);
}}}, {entityTypes:["minecraft:player"]});